<?php
/**
 * Useful class for debugging.
 */

namespace Flat101\Security\Util;

use Error;
use Exception;
use WP_Error;

use const Flat101\Security\PREFIX;

class Debug {

	const CHRONO_START = PREFIX . '__chrono_start';

	/**
	 * Write to the log.
	 *
	 * @param string $message Text to write in the log.
	 * @return void
	 */
	protected static function error_log( $message ) {

		// phpcs:disable WordPress.PHP.DevelopmentFunctions.error_log_error_log
		if ( defined( 'FLAT101_SECURITY__DEBUG_PATH' ) && '' !== FLAT101_SECURITY__DEBUG_PATH ) {
			// [15-Jun-2021 13:22:17 UTC] xxx
			$txt = '[' . gmdate( 'd-M-Y H:i:s \U\T\C' ) . '] ' . $message . "\n";
			error_log( $txt, 3, FLAT101_SECURITY__DEBUG_PATH );
		} else {
			error_log( $message );
		}
		// phpcs:enable WordPress.PHP.DevelopmentFunctions.error_log_error_log
	}

	/**
	 * Designed to write any type of data to the log.
	 *
	 * If WP_DEBUG is on write to:
	 *     /wp-content/debug.log
	 *
	 * @param mixed ...$items As many parameters as you need of whatever type.
	 * @return void
	 */
	public static function debug( ...$items ) {

		if ( defined( 'WP_DEBUG' ) && ( true === WP_DEBUG ) ) {
			foreach ( $items as $item ) {
				if ( $item instanceof Exception || $item instanceof Error ) {
					static::error_log( (string) $item );
					continue;
				}

				if ( $item instanceof WP_Error ) {
					$item = array(
						'errors' => $item->errors,
						'data'   => $item->get_all_error_data(),
					);
				}

				static::error_log(
					(string) wp_json_encode(
						$item,
						JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT
					)
				);
			}
		}
	}

	/**
	 * Write the text of the message with optional data.
	 *
	 * @param string $message Text to write in the log.
	 * @param mixed  ...$data Dataset related to this message. As many parameters as you need.
	 * @return void
	 */
	public static function log( $message, ...$data ) {

		if ( defined( 'WP_DEBUG' ) && ( true === WP_DEBUG ) ) {
			static::error_log( $message );

			if ( ! empty( $data ) ) {
				static::debug( ...$data );
			}
		}
	}

	/**
	 * Write to the log the last query to the database.
	 *
	 * @return void
	 */
	public static function last_query() {

		/** @var \wpdb */
		global $wpdb;

		$last_query = preg_replace( '/\s+/', ' ', $wpdb->last_query );

		static::debug( $last_query );
	}

	/**
	 * Write to the log all the queries to the database until now.
	 *
	 * Require:
	 *   define( 'SAVEQUERIES', true );
	 *
	 * @return void
	 */
	public static function queries() {

		/** @var \wpdb */
		global $wpdb;

		static::debug( $wpdb->queries );
	}


	/**
	 * Take the current instant as the start of the chrono.
	 *
	 * @return void
	 */
	public static function chrono_start() {

		// phpcs:ignore WordPress.NamingConventions.PrefixAllGlobals.NonPrefixedVariableFound
		$GLOBALS[ self::CHRONO_START ] = microtime( true );
	}

	/**
	 * Writes to the log the time elapsed since the last time 'chrono_start()' was called.
	 * If 'chrono_start()' has not been called previously it will not write anything.
	 *
	 * @return void
	 */
	public static function chrono_end() {

		if ( isset( $GLOBALS[ self::CHRONO_START ] ) ) {
			$chrono_end = microtime( true );
			static::log( 'Time: ' . round( ( $chrono_end - $GLOBALS[ self::CHRONO_START ] ) * 1000, 0 ) . ' ms' );
		}
	}

	/**
	 * Return human-readable information about a data for debugging.
	 *
	 * @param mixed  $data Whatever you want.
	 * @param string $prefix Prefix text.
	 * @param string $sufix Sufix text.
	 * @return string
	 */
	public static function to_string( $data, $prefix = "\n", $sufix = "\n" ) {

		if ( $data instanceof WP_Error ) {
			$data = array(
				'errors' => $data->errors,
				'data'   => $data->get_all_error_data(),
			);
		}

		// phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
		return $prefix . wp_json_encode( $data, JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT ) . $sufix . "\n";
	}


	/**
	 * Echo any type of data to the screen (only if WP_DEBUG is on).
	 *
	 * phpcs:disable PHPCompatibility.Keywords.ForbiddenNames.echoFound
	 *
	 * @param mixed  $data Whatever you want.
	 * @param string $prefix Prefix text.
	 * @param string $sufix Sufix text.
	 * @return void
	 */
	public static function echo( $data, $prefix = "\n", $sufix = "\n" ) {

		// phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
		echo static::to_string( $data, $prefix, $sufix );
	}

	/**
	 * Echo any type of data to the screen surrounded by <pre> (only if WP_DEBUG is on).
	 *
	 * @param mixed $data Whatever you want.
	 * @return void
	 */
	public static function pre( $data ) {

		static::echo( $data, "<pre>\n", "</pre>\n" );
	}

	/**
	 * Write to log the current request.
	 *
	 * @return void
	 */
	public static function request() {

		if (
			defined( 'WP_DEBUG' ) &&
			( true === WP_DEBUG ) &&
			defined( 'FLAT101_SECURITY__DEBUG_REQUESTS' ) &&
			( true === FLAT101_SECURITY__DEBUG_REQUESTS ) &&
			! empty( $_SERVER['REQUEST_URI'] ) &&
			// phpcs:ignore WordPress.Security.NonceVerification.Missing
			! ( isset( $_POST['action'] ) && 'heartbeat' === $_POST['action'] )
		) {
			static::log( '-----------------------------------------------------------------' );

			$info = array(
				isset( $_SERVER['REQUEST_METHOD'] ) ? sanitize_text_field( wp_unslash( $_SERVER['REQUEST_METHOD'] ) ) : null,
				isset( $_SERVER['REQUEST_URI'] ) ? esc_url_raw( wp_unslash( $_SERVER['REQUEST_URI'] ) ) : null,
			);
			static::log( implode( ' ', $info ) );

			// phpcs:disable WordPress.Security.NonceVerification.Recommended
			if ( isset( $_REQUEST['action'] ) ) {
				static::log( 'ACTION: ' . sanitize_text_field( wp_unslash( $_REQUEST['action'] ) ) );
			}
			// phpcs:enable WordPress.Security.NonceVerification.Recommended

			// phpcs:disable WordPress.Security.NonceVerification.Missing
			if ( count( $_POST ) > 0 ) {
				// phpcs:ignore WordPress.PHP.DevelopmentFunctions.error_log_print_r
				static::log( 'POST: ' . print_r( $_POST, true ) );
			}
			// phpcs:enable WordPress.Security.NonceVerification.Missing
		}
	}
}
